/*
 * Copyright 2018 Confluent Inc.
 *
 * Licensed under the Confluent Community License (the "License"); you may not use
 * this file except in compliance with the License.  You may obtain a copy of the
 * License at
 *
 * http://www.confluent.io/confluent-community-license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 */

package io.confluent.ksql.parser.tree;

import static java.util.Locale.ENGLISH;
import static java.util.Objects.requireNonNull;

import com.google.common.base.Preconditions;
import java.util.Objects;
import java.util.Optional;

public class BooleanLiteral
    extends Literal {

  private final boolean value;

  public BooleanLiteral(final String value) {
    this(Optional.empty(), value);
  }

  public BooleanLiteral(final NodeLocation location, final String value) {
    this(Optional.of(location), value);
  }

  private BooleanLiteral(final Optional<NodeLocation> location, final String value) {
    super(location);
    requireNonNull(value, "value is null");
    Preconditions.checkArgument(
        value.toLowerCase(ENGLISH).equals("true") || value.toLowerCase(ENGLISH).equals("false"));

    this.value = value.toLowerCase(ENGLISH).equals("true");
  }

  public boolean getValue() {
    return value;
  }

  @Override
  public <R, C> R accept(final AstVisitor<R, C> visitor, final C context) {
    return visitor.visitBooleanLiteral(this, context);
  }

  @Override
  public int hashCode() {
    return Objects.hash(value);
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    final BooleanLiteral other = (BooleanLiteral) obj;
    return Objects.equals(this.value, other.value);
  }
}
